//
//  kde.cpp
//  KDE
//
//  Created by Kento Kodama on 2/23/15.
//  Copyright (c) 2015 Saigo Laboratoire. All rights reserved.
//
//  This is free software with ABSOLUTELY NO WARRANTY.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cfloat>
#include <ctime>
#include <getopt.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <map>
#include <new>
#include <random>
#include <algorithm>
#include "Eigen/Core"
#include "Eigen/LU"
#include "Eigen/Sparse"
#include "Eigen/QR"
#include "Eigen/Dense"

#include "MatrixExt.h"
#include "data.h"
#include "krr.h"

using namespace std;
using namespace Eigen;
using namespace EigenExt;

//static int qc = 1;

enum{
     NotShuffle = 0,
     RandShuffle = 1,
     NRandShuffle = 2,
     OptShuffle  = 3
};

typedef unsigned int SHUFFLE;

void Error() {
     cout <<
             "Usage: kde <[Option] ...>\n\n"
             //"       --qt / --cc                : quantitative or case-control(default=qt)\n"
             "       --gen    / -g [filename]:  genotype data\n"
             "       --phe    / -p [filename]:  phenotype data\n"
             "       --out    / -o [filename]:  output filename (default: kde.result)\n"
             "       --degree / -d [int]     :  number of interactions (default=10)\n"
             "       --fold   / -f [int]     :  number of fold (default=4)\n\n"
             "Example: kde --genotype test.gen --phenotype test.phe --degree 5\n"
             "         kde -g test.gen -p test.phe -d 5\n"
             << endl;
     exit(1);
}

void Signature() {
     cout <<
             "\n"
             "@----------------------------------------------------------@\n"
             "|       kde       |       ver 1.01       |   23/Feb/2015   |\n"
             "|----------------------------------------------------------|\n"
             "|  (C) 2015 Kento Kodama, GNU General Public Licence, v2   |\n"
             "|----------------------------------------------------------|\n"
             "| For documentation, criterion and bug-report instruction: |\n"
             "|       http://saigo-www.bio.kyutech.ac.jp/~kodama/        |\n"
             "@----------------------------------------------------------@\n"
             << endl;
}

int main(int argc, char* argv[]) {
     //////////////////////////////////////////////////////////////////////////
     //preparation
     if ( argc < 2 ) {
          Error();
          exit(1);
     }

     int c;
     unsigned int fold = 4;
     Genotypes Gen;
     Phenotypes Phe;
     KernelRidgeRegression KRR;
     SHUFFLE sh = RandShuffle;
     ofstream log("kde.log");
     string outfilename = "kde.result",gen_filename, phe_filename;
     bool gen_load = false, phe_load = false, degree_num = false;

     Signature();
     time_t curr = time(0);
     string tdstamp = (string) ctime(&curr);
     cout << "Writing text to log file:  [ kde.log ]" << endl;
     log << "Writing text to log file:  [ kde.log ]" << endl;
     cout << "Started analysis:  " << tdstamp << endl;
     log << "Started analysis:  " << tdstamp << endl;
     cout << "Options in effect:" << endl;
     log << "Options in effect:" << endl;

     while ( 1 ) {
          static struct option long_options[] = {
               //{"qt", no_argument, &qc, 1},
               //{"cc", no_argument, &qc, 0}, 
               {"gen", required_argument, 0, 'g'},
               {"phe", required_argument, 0, 'p'},
               {"degree", required_argument, 0, 'd'},
               {"fold", required_argument, 0, 'f'},
               {"out", required_argument, 0, 'o'},
	     {"shuffle", required_argument, 0, 's'},
               {"help", no_argument, 0, 'h'},
               {0, 0, 0, 0}
          };

          int option_index = 0;

          c = getopt_long(argc, argv, "g:p:d:f:o:s:h", long_options, &option_index);
          if ( c == -1 ) break;

          switch (c) {
               case 0:
                    /* If this option set a flag, do nothing else now. */
                    if ( long_options[option_index].flag != 0 ) break;
                    printf("option %s", long_options[option_index].name);
                    if ( optarg ) printf(" with arg %s", optarg);
                    printf("\n");
                    break;

               case 'g':
                    printf("\t--gen %s\n", optarg);
                    log << "\t--gen " << optarg << endl;
                    gen_filename = optarg;
                    gen_load = Gen.LoadX(optarg);
                    if ( !gen_load ) exit(1);
                    break;

               case 'p':
                    printf("\t--phe %s\n", optarg);
                    log << "\t--phe" << optarg << endl;
                    phe_filename = optarg;
                    phe_load = Phe.LoadY(optarg);
                    if ( !phe_load ) exit(1);
                    break;

               case 'd':
                    printf("\t--degree %s\n", optarg);
                    log << "\t--degree " << optarg << endl;
                    degree_num = KRR.SetDegree(atoi(optarg));
                    break;

               case 'f':
                    printf("\t--fold %s\n", optarg);
                    log << "\t--fold " << optarg << endl;
                    fold = atoi(optarg);
                    break;

               case 'o':
                    printf("\t--out %s\n", optarg);
                    log << "\t--out " << optarg << endl;
                    outfilename = optarg;
                    break;

	       case 's':
		    printf("\t--shuffle %s\n",optarg);
		    log << "\t--shuffle " << optarg << endl;
		    sh = atoi(optarg);
		    break;

               case '?':
                    /* getopt_long already printed an error message. */
                    break;

               case 'h':
                    Error();
                    break;

               default:
                    Error();
                    exit(1);
          }
     }

     if ( optind < argc ) {
          printf("non-option ARGV-elements: ");
          while ( optind < argc ) printf("%s ", argv[optind++]);
          putchar('\n');
     }
     
     ofstream res(outfilename.c_str());
     res << "degree\tRSS\ttuning parameter" << endl;
     
     //////////////////////////////////////////////////////////////////////////
     // pretreatment of genotype and phenotype data
     stringstream ind_info, snp_info, phe_info;
     Gen.SetN();
     ind_info << Gen.GetN() << " individuals read from [ " << gen_filename << " ]" << endl;
     Gen.SetP();
     snp_info << Gen.GetP() << " SNPs read from [ " << gen_filename << " ]" << endl;
     Phe.SetN();
     phe_info << Phe.GetN() << " individuals read from [ " << phe_filename << " ]" << endl;
     cout << endl << ind_info.str() << snp_info.str() << phe_info.str();
     log << endl << ind_info.str() << snp_info.str() << phe_info.str();
     
     if ( Gen.GetN() != Phe.GetN() ){
          stringstream error_message;
          error_message << endl << "Number of individuals from [ " << gen_filename << " ] is not equal to that from [ " << phe_filename << " ]." << endl;
          cout << error_message.str() << endl;
          log << error_message.str() << endl;
          exit(1);
     }
     Phe.Centred();
     KRR.SetFold(fold, Gen.GetN());
     if (!degree_num) KRR.SetDegree(10);
     KRR.SetTuningParameter();
     
     unsigned int comp_num = KRR.GetFold() - Gen.GetN()%KRR.GetFold(), row;
     if (Gen.GetN()%KRR.GetFold() != 0){
          mt19937 engine;
          uniform_int_distribution<int> distribution( 1, Gen.GetN() );
          for (int i = 0; i < comp_num; ++i){
               row = distribution(engine);
               Gen.ComplementX(row);
               Gen.SetN();
               Phe.ComplementY(row);
               Phe.SetN();
          }
          cout << "Complemented data for Cross-Validation" << endl;
          log << "Complemented data for Cross-Validation" << endl;
     }
     
     // data shuffle 
     vector<unsigned int> order;
     if ( sh != NotShuffle ){
          if ( sh == RandShuffle || sh == NRandShuffle){
               for (int i = 0; i < Gen.GetN(); ++i) order.push_back(i);
	       if ( sh == RandShuffle ) srand(1234);
	       else srand(time(NULL));
	       random_shuffle(order.begin(), order.end());
          }
          else if ( sh == OptShuffle ) order = Phe.SortIndexByY(KRR.GetFold());
          else exit(1);
          Gen.ShuffleX(order);
          Phe.ShuffleY(order);
     }

     //////////////////////////////////////////////////////////////////////////
     // SNP-SNP Interaction Estimation by Kernel Ridge Regression
     KRR.SetNumber(Gen.GetN(), KRR.GetFold());
     KRR.SetOnesVector();
     KRR.SetIdentityMatrix();
     KRR.SetRSSMatrix();
     
     for (int i = 0; i < KRR.GetFold(); ++i){
          KRR.SetKernel(Gen.GetX(),CVTYPETrain, i+1);
          KRR.SetKernel(Gen.GetX(),CVTYPEValid, i+1);
          KRR.SetKernel(Gen.GetX(),CVTYPETest, i+1);
          KRR.SetY(Phe.GetY(),CVTYPETrain, i+1);
          KRR.SetY(Phe.GetY(),CVTYPEValid, i+1);
          KRR.SetY(Phe.GetY(),CVTYPETest, i+1);
          KRR.KernelNormalization();
          res << KRR.Regression() << endl;
     }
     cout << endl;
     cout << "Writing predicted results to [ " << outfilename << " ]" << endl;
     log << "Writing predicted results to [ " << outfilename << " ]" << endl;

     tdstamp = (string) ctime(&curr);
     cout << endl << "Analysis finished:  " << tdstamp << endl;
     log << endl << "Analysis finished:  " << tdstamp << endl;

     return 0;
}
