//
//  krr.h
//  KDE
//
//  Created by Kento Kodama on 2/23/15.
//  Copyright (c) 2015 Saigo Laboratoire. All rights reserved.
//
//  This is free software with ABSOLUTELY NO WARRANTY.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cfloat>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <new>
#include "Eigen/Core"
#include "Eigen/LU"
#include "Eigen/Sparse"
#include "Eigen/QR"
#include "Eigen/Dense"

#include "MatrixExt.h"

using namespace std;
using namespace Eigen;
using namespace EigenExt;

enum{
     CVTYPETrain = 1,
     CVTYPEValid = 1 << 1,
     CVTYPETest  = 1 << 2
};

typedef unsigned int CVTYPE;

class KernelRidgeRegression {
private:
     MatrixXd train_K;
     MatrixXd valid_K;
     MatrixXd test_K;
     MatrixXd train_Y;
     MatrixXd valid_Y;
     MatrixXd test_Y;
     MatrixXd alpha;
     MatrixXd valid_RSS;
     MatrixXd test_RSS;
     MatrixXd ones;
     MatrixXd I;
     unsigned int train_num;
     unsigned int valid_num;
     unsigned int test_num;
     unsigned int fold;                       //number of fold for Cross-Validation
     unsigned int degree;                     //maximum number of SNPs in interaction
     vector<double> TuningParameter; //tuning parameter for ridge regression
     vector<double> KinshipTrain;
     vector<double> KinshipValid;
     vector<double> KinshipTest;
     
public:
     virtual bool SetFold(unsigned int fold, unsigned int num);
     virtual unsigned int GetFold();
     virtual bool SetDegree(unsigned int degree);
     virtual unsigned int GetDegree();
     virtual void SetTuningParameter();
     virtual double GetTuningParameter(unsigned int i);
     virtual void SetNumber(unsigned int n, unsigned int fold);
     virtual void SetKernel(MatrixXd& X, CVTYPE cv, unsigned int fold_num);
     virtual void SetY(MatrixXd& Y, CVTYPE cv, unsigned int fold_num);
     virtual void SetOnesVector();
     virtual void SetIdentityMatrix();
     virtual void SetRSSMatrix();     
     virtual void KernelNormalization();
     virtual string Regression();
};
