//
//  krr.cpp
//  KDE
//
//  Created by Kento Kodama on 2/23/15.
//  Copyright (c) 2015 Saigo Laboratoire. All rights reserved.
//
//  This is free software with ABSOLUTELY NO WARRANTY.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cfloat>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <new>
#include "Eigen/Core"
#include "Eigen/LU"
#include "Eigen/Sparse"
#include "Eigen/QR"
#include "Eigen/Dense"
#include "MatrixExt.h"
#include "krr.h"

using namespace std;
using namespace Eigen;
using namespace EigenExt;

bool KernelRidgeRegression::SetFold(unsigned int fold,  unsigned int num) {
     if ( !(fold > 3  && fold < num + 1)){
          printf("\nPlease set the number of folds from four to %d.\n\n", num);
          exit(1);
     }
     else this->fold = fold;
     return true;
}

bool KernelRidgeRegression::SetDegree(unsigned int degree) {
     if ( degree < 1 ){
          cout << endl << "Please set the number of degree one or more.\n" << endl << endl;
          exit(1);
     }
     else this->degree = degree;
     return true;
}

void KernelRidgeRegression::SetTuningParameter() {
     this->TuningParameter.push_back(0.001);
     this->TuningParameter.push_back(0.01);
     this->TuningParameter.push_back(0.1);
     this->TuningParameter.push_back(1);
     this->TuningParameter.push_back(10);
     this->TuningParameter.push_back(100);
}

unsigned int KernelRidgeRegression::GetFold(void) {
     return (fold);
}

unsigned int KernelRidgeRegression::GetDegree(void) {
     return (degree);
}

double KernelRidgeRegression::GetTuningParameter(unsigned int i) {
     return (TuningParameter[i - 1]);
}

void KernelRidgeRegression::SetNumber(unsigned int n, unsigned int fold) {
     this->train_num = n * (fold - 2) / fold;
     this->valid_num = n / fold;
     this->test_num = n / fold;
}

void KernelRidgeRegression::SetKernel(MatrixXd& X, CVTYPE cv, unsigned int fold_num) {
     int n = X.rows();
     int cols = X.cols();

     MatrixXd train_X(this->train_num, cols);
     if ( fold_num < 4 ) train_X = X.block(n * (fold_num - 1) / this->fold, 0, this->train_num, cols);
     else {
          train_X.block(0, 0, n * (this->fold - fold_num + 1) / this->fold, cols) = X.block(n * (fold_num - 1) / this->fold, 0, n * (this->fold - fold_num + 1) / this->fold, cols);
          train_X.block(n * (this->fold - fold_num + 1) / this->fold, 0, n * (fold_num - 3) / this->fold, cols) = X.block(0, 0, n * (fold_num - 3) / this->fold, cols);
     }

     if ( cv == CVTYPETrain ) {
          this->train_K.resize(this->train_num, this->train_num);
          //cout << this->train_K << endl << endl << train_X << endl << endl;
          this->train_K = train_X * train_X.transpose() + this->ones;
          for (int i = 0;i < this->train_num; ++i) this->KinshipTrain.push_back(this->train_K(i,i));
     } else if ( cv == CVTYPEValid ) {
          MatrixXd valid_X(this->valid_num, cols), valid(this->valid_num, this->valid_num);
          if ( fold_num < 3 ) valid_X = X.block(n * (this->fold + fold_num - 3) / fold, 0, this->valid_num, cols);
          else valid_X = X.block(n * (fold_num - 3) / this->fold, 0, this->valid_num, cols);
          this->valid_K = valid_X * train_X.transpose() + (this->ones).block(0, 0, this->valid_num, this->train_num);
          //cout << this->valid_K << endl << endl << valid_X << endl << endl;
          valid = valid_X * valid_X.transpose()+ (this->ones).block(0, 0, this->valid_num, this->valid_num);
          for (int i = 0;i < this->valid_num; ++i) this->KinshipValid.push_back(valid(i,i));
     } else {
          MatrixXd test_X(this->test_num, cols), test(this->test_num, this->test_num);
          if ( fold_num == 1 ) test_X = X.block(n * (this->fold - 1) / this->fold, 0, this->test_num, cols);
          else test_X = X.block(n * (fold_num - 2) / this->fold, 0, this->test_num, cols);
          this->test_K = test_X * train_X.transpose() + (this->ones).block(0, 0, this->test_num, this->train_num);
          //cout << this->test_K << endl << endl << test_X << endl << endl;
          test = test_X * test_X.transpose() + (this->ones).block(0, 0, this->test_num, this->test_num);
          for (int i = 0;i < this->test_num; ++i) this->KinshipTest.push_back(test(i,i));
     }
}

void KernelRidgeRegression::SetY(MatrixXd& Y, CVTYPE cv, unsigned int fold_num) {
     int n = Y.rows();
     if ( cv == CVTYPETrain ) {
          if ( fold_num < 4 ) this->train_Y = Y.block(n * (fold_num - 1) / this->fold, 0, this->train_num, 1);
          else {
               this->train_Y.block(0, 0, n * (this->fold - fold_num + 1) / this->fold, 1) = Y.block(n * (fold_num - 1) / this->fold, 0, n * (this->fold - fold_num + 1) / this->fold, 1);
               this->train_Y.block(n * (this->fold - fold_num + 1) / this->fold, 0, n * (fold_num - 3) / this->fold, 1) = Y.block(0, 0, n * (fold_num - 3) / this->fold, 1);
          }
          //          cout << this->train_Y << endl << endl;
     } else if ( cv == CVTYPEValid ) {
          if ( fold_num < 3 ) this->valid_Y = Y.block(n * (this->fold + fold_num - 3) / this->fold, 0, this->valid_num, 1);
          else this->valid_Y = Y.block(n * (fold_num - 3) / this->fold, 0, this->valid_num, 1);
          //          cout << this->valid_Y << endl << endl;
     } else {
          if ( fold_num == 1 ) this->test_Y = Y.block(n * (this->fold - 1) / this->fold, 0, this->test_num, 1);
          else this->test_Y = Y.block(n * (fold_num - 2) / this->fold, 0, this->test_num, 1);
          //          cout << this->test_Y << endl << endl;
     }
}

void KernelRidgeRegression::SetOnesVector() {
     this->ones.resize(this->train_num, this->train_num);
     this->ones << MatrixXd::Ones(this->train_num, this->train_num);
}

void KernelRidgeRegression::SetIdentityMatrix() {
     this->I.resize(this->train_num, this->train_num);
     this->I << MatrixXd::Identity(this->train_num, this->train_num);
}

void KernelRidgeRegression::SetRSSMatrix() {
     (this->valid_RSS).resize(this->degree + 1, (this->TuningParameter).size());
     (this->test_RSS).resize(this->degree + 1, (this->TuningParameter).size());
}

void KernelRidgeRegression::KernelNormalization() {
     //train standardization
     for (int i = 0; i < this->train_num; ++i) {
        for (int j = 0; j < this->train_num; ++j) {
             if ( j > i - 1 ) {
                  this->train_K(i, j) = this->train_K(i, j) / sqrt(this->KinshipTrain[i] * this->KinshipTrain[j]);
                  this->train_K(j, i) = this->train_K(i, j);
             }
        }
     }
     
     //validation astandardization
     for (int i = 0; i < this->valid_num; ++i) {
        for (int j = 0; j < this->train_num; ++j) {
             this->valid_K(i, j) = this->valid_K(i, j) / sqrt(this->KinshipValid[i] * this->KinshipTrain[j]);
        }
     }
     
     //test standardization
     for (int i = 0; i < this->test_num; ++i) {
        for (int j = 0; j < this->train_num; ++j) {
             this->test_K(i, j) = this->test_K(i, j) / sqrt(this->KinshipTest[i] * this->KinshipTrain[j]);
        }
     }
}

string KernelRidgeRegression::Regression() {
     static int count;
     for (int i = 0; i < this->degree + 1; ++i) {
          for (int j = 0; j < (this->TuningParameter).size(); ++j) {
               if ( i == 0 ) {
                    this->alpha = (this->TuningParameter[j] * this->I).colPivHouseholderQr().solve(this->train_Y);
                    this->valid_RSS(i, j) = (this->valid_Y).array().pow(2.0).sum();
                    this->test_RSS(i, j) = (this->test_Y).array().pow(2.0).sum();
               } else {
                    this->alpha = ((MatrixXd)(this->train_K.array().pow(i)) + this->TuningParameter[j] * this->I).colPivHouseholderQr().solve(this->train_Y);
                    this->valid_RSS(i, j) = (((MatrixXd)(this->valid_K.array().pow(i)) * this->alpha - this->valid_Y).array().pow(2.0)).sum();
                    this->test_RSS(i, j) = (((MatrixXd)(this->test_K.array().pow(i)) * this->alpha - this->test_Y).array().pow(2.0)).sum();
               }
               ++count;
               cout << "Performing tests: " << int(100 * double(count)/((this->degree + 1) * ((this->TuningParameter).size()) * (this->fold))) << " %";
               if ( count < ((this->degree + 1) * ((this->TuningParameter).size()) * (this->fold))) {
                    cout << "\r";
                    cout.flush();
               }
          }
     }

     MatrixXf::Index test_minRow, test_minCol;
     this->valid_RSS.minCoeff(&test_minRow, &test_minCol);
     
     // output result
     stringstream res;
     res << test_minRow << "\t" << this->test_RSS(test_minRow, test_minCol)/this->test_num << "\t" << this->TuningParameter[test_minCol];

     return (res.str()); //test_RSS, number of SNPs in interaction and tuning parameter
}
